use std::error::Error;

use associations::BelongsTo;
use backend::Backend;
use deserialize::{self, FromSqlRow, Queryable, QueryableByName};
use expression::{AppearsOnTable, AsExpression, AsExpressionList, Expression, NonAggregate,
                 SelectableExpression};
use insertable::{CanInsertInSingleQuery, InsertValues, Insertable};
use query_builder::*;
use query_source::*;
use result::QueryResult;
use row::*;
use sql_types::{HasSqlType, NotNull};
use util::TupleAppend;

macro_rules! tuple_impls {
    ($(
        $Tuple:tt {
            $(($idx:tt) -> $T:ident, $ST:ident, $TT:ident,)+
    }
    )+) => {
        $(
            impl<$($T),+, __DB> HasSqlType<($($T,)+)> for __DB where
                $(__DB: HasSqlType<$T>),+,
                __DB: Backend,
            {
                fn metadata(_: &__DB::MetadataLookup) -> __DB::TypeMetadata {
                    unreachable!("Tuples should never implement `ToSql` directly");
                }

                fn row_metadata(out: &mut Vec<__DB::TypeMetadata>, lookup: &__DB::MetadataLookup) {
                    $(<__DB as HasSqlType<$T>>::row_metadata(out, lookup);)+
                }
            }

            impl<$($T),+> NotNull for ($($T,)+) {
            }

            impl<$($T),+, $($ST),+, __DB> FromSqlRow<($($ST,)+), __DB> for ($($T,)+) where
                __DB: Backend,
                $($T: FromSqlRow<$ST, __DB>),+,
            {
                const FIELDS_NEEDED: usize = $($T::FIELDS_NEEDED +)+ 0;

                fn build_from_row<RowT: Row<__DB>>(row: &mut RowT) -> Result<Self, Box<Error+Send+Sync>> {
                    Ok(($(try!($T::build_from_row(row)),)+))
                }
            }

            impl<$($T),+, $($ST),+, __DB> Queryable<($($ST,)+), __DB> for ($($T,)+) where
                __DB: Backend,
                $($T: Queryable<$ST, __DB>),+,
            {
                type Row = ($($T::Row,)+);

                fn build(row: Self::Row) -> Self {
                    ($($T::build(row.$idx),)+)
                }
            }

            impl<$($T,)+ __DB> QueryableByName<__DB> for ($($T,)+)
            where
                __DB: Backend,
                $($T: QueryableByName<__DB>,)+
            {
                fn build<RowT: NamedRow<__DB>>(row: &RowT) -> deserialize::Result<Self> {
                    Ok(($($T::build(row)?,)+))
                }
            }

            impl<$($T: Expression + NonAggregate),+> Expression for ($($T,)+) {
                type SqlType = ($(<$T as Expression>::SqlType,)+);
            }

            impl<$($T: QueryFragment<__DB>),+, __DB: Backend> QueryFragment<__DB> for ($($T,)+) {
                #[allow(unused_assignments)]
                fn walk_ast(&self, mut out: AstPass<__DB>) -> QueryResult<()> {
                    let mut needs_comma = false;
                    $(
                        if !self.$idx.is_noop()? {
                            if needs_comma {
                                out.push_sql(", ");
                            }
                            self.$idx.walk_ast(out.reborrow())?;
                            needs_comma = true;
                        }
                    )+
                    Ok(())
                }
            }

            impl<$($T,)+ Tab> ColumnList for ($($T,)+)
            where
                $($T: ColumnList<Table = Tab>,)+
            {
                type Table = Tab;

                fn walk_ast<__DB: Backend>(&self, mut out: AstPass<__DB>) -> QueryResult<()> {
                    $(
                        if $idx != 0 {
                            out.push_sql(", ");
                        }
                        self.$idx.walk_ast(out.reborrow())?;
                    )+
                    Ok(())
                }
            }

            impl<$($T: QueryId),+> QueryId for ($($T,)+) {
                type QueryId = ($($T::QueryId,)+);

                const HAS_STATIC_QUERY_ID: bool = $($T::HAS_STATIC_QUERY_ID &&)+ true;
            }

            impl<$($T: Expression + NonAggregate),+> NonAggregate for ($($T,)+) {
            }

            impl<$($T,)+ Tab> UndecoratedInsertRecord<Tab> for ($($T,)+)
            where
                $($T: UndecoratedInsertRecord<Tab>,)+
            {
            }

            impl<$($T,)+ __DB> CanInsertInSingleQuery<__DB> for ($($T,)+)
            where
                __DB: Backend,
                $($T: CanInsertInSingleQuery<__DB>,)+
            {
                fn rows_to_insert(&self) -> Option<usize> {
                    $(debug_assert_eq!(self.$idx.rows_to_insert(), Some(1));)+
                    Some(1)
                }
            }

            impl<$($T,)+ $($ST,)+ Tab> Insertable<Tab> for ($($T,)+)
            where
                $($T: Insertable<Tab, Values = ValuesClause<$ST, Tab>>,)+
            {
                type Values = ValuesClause<($($ST,)+), Tab>;

                fn values(self) -> Self::Values {
                    ValuesClause::new(($(self.$idx.values().values,)+))
                }
            }

            impl<'a, $($T,)+ Tab> Insertable<Tab> for &'a ($($T,)+)
            where
                ($(&'a $T,)+): Insertable<Tab>,
            {
                type Values = <($(&'a $T,)+) as Insertable<Tab>>::Values;

                fn values(self) -> Self::Values {
                    ($(&self.$idx,)+).values()
                }
            }

            #[allow(unused_assignments)]
            impl<$($T,)+ Tab, __DB> InsertValues<Tab, __DB> for ($($T,)+)
            where
                Tab: Table,
                __DB: Backend,
                $($T: InsertValues<Tab, __DB>,)+
            {
                fn column_names(&self, mut out: AstPass<__DB>) -> QueryResult<()> {
                    let mut needs_comma = false;
                    $(
                        let noop_element = self.$idx.is_noop()?;
                        if !noop_element {
                            if needs_comma {
                                out.push_sql(", ");
                            }
                            self.$idx.column_names(out.reborrow())?;
                            needs_comma = true;
                        }
                    )+
                    Ok(())
                }
            }

            impl<$($T,)+ QS> SelectableExpression<QS> for ($($T,)+) where
                $($T: SelectableExpression<QS>,)+
                ($($T,)+): AppearsOnTable<QS>,
            {
            }

            impl<$($T,)+ QS> AppearsOnTable<QS> for ($($T,)+) where
                $($T: AppearsOnTable<QS>,)+
                ($($T,)+): Expression,
            {
            }

            impl<Target, $($T,)+> AsChangeset for ($($T,)+) where
                $($T: AsChangeset<Target=Target>,)+
                Target: QuerySource,
            {
                type Target = Target;
                type Changeset = ($($T::Changeset,)+);

                fn as_changeset(self) -> Self::Changeset {
                    ($(self.$idx.as_changeset(),)+)
                }
            }

            impl<$($T,)+ Parent> BelongsTo<Parent> for ($($T,)+) where
                A: BelongsTo<Parent>,
            {
                type ForeignKey = A::ForeignKey;
                type ForeignKeyColumn = A::ForeignKeyColumn;

                fn foreign_key(&self) -> Option<&Self::ForeignKey> {
                    self.0.foreign_key()
                }

                fn foreign_key_column() -> Self::ForeignKeyColumn {
                    A::foreign_key_column()
                }
            }

            impl<$($T,)+ Next> TupleAppend<Next> for ($($T,)+) {
                type Output = ($($T,)+ Next);

                #[allow(non_snake_case)]
                fn tuple_append(self, next: Next) -> Self::Output {
                    let ($($T,)+) = self;
                    ($($T,)+ next)
                }
            }

            impl<$($T,)+ ST> AsExpressionList<ST> for ($($T,)+) where
                $($T: AsExpression<ST>,)+
            {
                type Expression = ($($T::Expression,)+);

                fn as_expression_list(self) -> Self::Expression {
                    ($(self.$idx.as_expression(),)+)
                }
            }
        )+
    }
}

tuple_impls! {
    1 {
        (0) -> A, SA, TA,
    }
    2 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
    }
    3 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
    }
    4 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
    }
    5 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
    }
    6 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
    }
    7 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
    }
    8 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
    }
    9 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
    }
    10 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
    }
    11 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
    }
    12 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
    }
    13 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
    }
    14 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
    }
    15 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
    }
    16 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
    }
}

#[cfg(feature = "x32-column-tables")]
tuple_impls! {
    17 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
    }
    18 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
    }
    19 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
    }
    20 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
    }
    21 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
    }
    22 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
    }
    23 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
    }
    24 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
    }
    25 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
    }
    26 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
    }
    27 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
    }
    28 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
    }
    29 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
    }
    30 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
    }
    31 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
    }
    32 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
    }
}

#[cfg(feature = "x64-column-tables")]
tuple_impls! {
    33 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
    }
    34 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
    }
    35 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
    }
    36 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
    }
    37 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
    }
    38 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
    }
    39 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
    }
    40 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
    }
    41 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
    }
    42 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
    }
    43 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
    }
    44 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
    }
    45 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
    }
    46 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
    }
    47 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
    }
    48 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
    }
    49 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
    }
    50 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
    }
    51 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
    }
    52 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
    }
    53 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
    }
    54 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
    }
    55 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
    }
    56 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
    }
    57 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
    }
    58 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
    }
    59 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
    }
    60 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
    }
    61 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
    }
    62 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
    }
    63 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
    }
    64 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
    }
}

#[cfg(feature = "x128-column-tables")]
tuple_impls! {
    65 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
    }
    66 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
    }
    67 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
    }
    68 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
    }
    69 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
    }
    70 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
    }
    71 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
    }
    72 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
    }
    73 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
    }
    74 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
    }
    75 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
    }
    76 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
    }
    77 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
    }
    78 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
    }
    79 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
    }
    80 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
    }
    81 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
    }
    82 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
    }
    83 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
    }
    84 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
    }
    85 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
    }
    86 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
    }
    87 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
    }
    88 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
    }
    89 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
    }
    90 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
    }
    91 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
    }
    92 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
    }
    93 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
    }
    94 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
    }
    95 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
    }
    96 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
    }
    97 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
    }
    98 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
    }
    99 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
    }
    100 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
    }
    101 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
    }
    102 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
    }
    103 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
    }
    104 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
    }
    105 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
    }
    106 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
    }
    107 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
    }
    108 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
    }
    109 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
    }
    110 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
    }
    111 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
    }
    112 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
    }
    113 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
    }
    114 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
    }
    115 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
    }
    116 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
    }
    117 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
    }
    118 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
    }
    119 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
        (118) -> DO, SDO, TDO,
    }
    120 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
        (118) -> DO, SDO, TDO,
        (119) -> DP, SDP, TDP,
    }
    121 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
        (118) -> DO, SDO, TDO,
        (119) -> DP, SDP, TDP,
        (120) -> DQ, SDQ, TDQ,
    }
    122 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
        (118) -> DO, SDO, TDO,
        (119) -> DP, SDP, TDP,
        (120) -> DQ, SDQ, TDQ,
        (121) -> DR, SDR, TDR,
    }
    123 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
        (118) -> DO, SDO, TDO,
        (119) -> DP, SDP, TDP,
        (120) -> DQ, SDQ, TDQ,
        (121) -> DR, SDR, TDR,
        (122) -> DS, SDS, TDS,
    }
    124 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
        (118) -> DO, SDO, TDO,
        (119) -> DP, SDP, TDP,
        (120) -> DQ, SDQ, TDQ,
        (121) -> DR, SDR, TDR,
        (122) -> DS, SDS, TDS,
        (123) -> DT, SDT, TDT,
    }
    125 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
        (118) -> DO, SDO, TDO,
        (119) -> DP, SDP, TDP,
        (120) -> DQ, SDQ, TDQ,
        (121) -> DR, SDR, TDR,
        (122) -> DS, SDS, TDS,
        (123) -> DT, SDT, TDT,
        (124) -> DU, SDU, TDU,
    }
    126 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
        (118) -> DO, SDO, TDO,
        (119) -> DP, SDP, TDP,
        (120) -> DQ, SDQ, TDQ,
        (121) -> DR, SDR, TDR,
        (122) -> DS, SDS, TDS,
        (123) -> DT, SDT, TDT,
        (124) -> DU, SDU, TDU,
        (125) -> DV, SDV, TDV,
    }
    127 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
        (118) -> DO, SDO, TDO,
        (119) -> DP, SDP, TDP,
        (120) -> DQ, SDQ, TDQ,
        (121) -> DR, SDR, TDR,
        (122) -> DS, SDS, TDS,
        (123) -> DT, SDT, TDT,
        (124) -> DU, SDU, TDU,
        (125) -> DV, SDV, TDV,
        (126) -> DW, SDW, TDW,
    }
    128 {
        (0) -> A, SA, TA,
        (1) -> B, SB, TB,
        (2) -> C, SC, TC,
        (3) -> D, SD, TD,
        (4) -> E, SE, TE,
        (5) -> F, SF, TF,
        (6) -> G, SG, TG,
        (7) -> H, SH, TH,
        (8) -> I, SI, TI,
        (9) -> J, SJ, TJ,
        (10) -> K, SK, TK,
        (11) -> L, SL, TL,
        (12) -> M, SM, TM,
        (13) -> N, SN, TN,
        (14) -> O, SO, TO,
        (15) -> P, SP, TP,
        (16) -> Q, SQ, TQ,
        (17) -> R, SR, TR,
        (18) -> S, SS, TS,
        (19) -> T, ST, TT,
        (20) -> U, SU, TU,
        (21) -> V, SV, TV,
        (22) -> W, SW, TW,
        (23) -> X, SX, TX,
        (24) -> Y, SY, TY,
        (25) -> Z, SZ, TZ,
        (26) -> AA, SAA, TAA,
        (27) -> AB, SAB, TAB,
        (28) -> AC, SAC, TAC,
        (29) -> AD, SAD, TAD,
        (30) -> AE, SAE, TAE,
        (31) -> AF, SAF, TAF,
        (32) -> AG, SAG, TAG,
        (33) -> AH, SAH, TAH,
        (34) -> AI, SAI, TAI,
        (35) -> AJ, SAJ, TAJ,
        (36) -> AK, SAK, TAK,
        (37) -> AL, SAL, TAL,
        (38) -> AM, SAM, TAM,
        (39) -> AN, SAN, TAN,
        (40) -> AO, SAO, TAO,
        (41) -> AP, SAP, TAP,
        (42) -> AQ, SAQ, TAQ,
        (43) -> AR, SAR, TAR,
        (44) -> AS, SAS, TAS,
        (45) -> AT, SAT, TAT,
        (46) -> AU, SAU, TAU,
        (47) -> AV, SAV, TAV,
        (48) -> AW, SAW, TAW,
        (49) -> AX, SAX, TAX,
        (50) -> AY, SAY, TAY,
        (51) -> AZ, SAZ, TAZ,
        (52) -> BA, SBA, TBA,
        (53) -> BB, SBB, TBB,
        (54) -> BC, SBC, TBC,
        (55) -> BD, SBD, TBD,
        (56) -> BE, SBE, TBE,
        (57) -> BF, SBF, TBF,
        (58) -> BG, SBG, TBG,
        (59) -> BH, SBH, TBH,
        (60) -> BI, SBI, TBI,
        (61) -> BJ, SBJ, TBJ,
        (62) -> BK, SBK, TBK,
        (63) -> BL, SBL, TBL,
        (64) -> BM, SBM, TBM,
        (65) -> BN, SBN, TBN,
        (66) -> BO, SBO, TBO,
        (67) -> BP, SBP, TBP,
        (68) -> BQ, SBQ, TBQ,
        (69) -> BR, SBR, TBR,
        (70) -> BS, SBS, TBS,
        (71) -> BT, SBT, TBT,
        (72) -> BU, SBU, TBU,
        (73) -> BV, SBV, TBV,
        (74) -> BW, SBW, TBW,
        (75) -> BX, SBX, TBX,
        (76) -> BY, SBY, TBY,
        (77) -> BZ, SBZ, TBZ,
        (78) -> CA, SCA, TCA,
        (79) -> CB, SCB, TCB,
        (80) -> CC, SCC, TCC,
        (81) -> CD, SCD, TCD,
        (82) -> CE, SCE, TCE,
        (83) -> CF, SCF, TCF,
        (84) -> CG, SCG, TCG,
        (85) -> CH, SCH, TCH,
        (86) -> CI, SCI, TCI,
        (87) -> CJ, SCJ, TCJ,
        (88) -> CK, SCK, TCK,
        (89) -> CL, SCL, TCL,
        (90) -> CM, SCM, TCM,
        (91) -> CN, SCN, TCN,
        (92) -> CO, SCO, TCO,
        (93) -> CP, SCP, TCP,
        (94) -> CQ, SCQ, TCQ,
        (95) -> CR, SCR, TCR,
        (96) -> CS, SCS, TCS,
        (97) -> CT, SCT, TCT,
        (98) -> CU, SCU, TCU,
        (99) -> CV, SCV, TCV,
        (100) -> CW, SCW, TCW,
        (101) -> CX, SCX, TCX,
        (102) -> CY, SCY, TCY,
        (103) -> CZ, SCZ, TCZ,
        (104) -> DA, SDA, TDA,
        (105) -> DB, SDB, TDB,
        (106) -> DC, SDC, TDC,
        (107) -> DD, SDD, TDD,
        (108) -> DE, SDE, TDE,
        (109) -> DF, SDF, TDF,
        (110) -> DG, SDG, TDG,
        (111) -> DH, SDH, TDH,
        (112) -> DI, SDI, TDI,
        (113) -> DJ, SDJ, TDJ,
        (114) -> DK, SDK, TDK,
        (115) -> DL, SDL, TDL,
        (116) -> DM, SDM, TDM,
        (117) -> DN, SDN, TDN,
        (118) -> DO, SDO, TDO,
        (119) -> DP, SDP, TDP,
        (120) -> DQ, SDQ, TDQ,
        (121) -> DR, SDR, TDR,
        (122) -> DS, SDS, TDS,
        (123) -> DT, SDT, TDT,
        (124) -> DU, SDU, TDU,
        (125) -> DV, SDV, TDV,
        (126) -> DW, SDW, TDW,
        (127) -> DX, SDX, TDX,
    }
}
